"""
The View classes render the data returned by a Source as a Panel
object.
"""

from io import StringIO

import numpy as np
import param
import panel as pn

from bokeh.models import NumeralTickFormatter

from ..sources import Source
from ..transforms import Transform


class View(param.Parameterized):
    """
    A View renders the data returned by a Source as a Viewable Panel
    object. The base class provides methods which query the Source for
    the latest data given the current filters and applies all
    specified `transforms`.

    Subclasses should use these methods to query the data and return
    a Viewable Panel object in the `get_panel` method.
    """

    application = param.Parameter(constant=True, doc="""
        The containing application.""")

    filters = param.List(constant=True, doc="""
        A list of Filter object providing the query parameters for the
        Source.""")

    monitor = param.Parameter(constant=True, doc="""
        The Monitor class which owns this view.""")

    source = param.ClassSelector(class_=Source, constant=True, doc="""
        The Source to query for the data.""")

    transforms = param.List(constant=True, doc="""
        A list of transforms to apply to the data returned by the
        Source before visualizing it.""")

    table = param.String(doc="The table being visualized.")

    field = param.String(doc="The field being visualized.")

    view_type = None

    def __init__(self, **params):
        self.kwargs = {k: v for k, v in params.items() if k not in self.param}
        super().__init__(**{k: v for k, v in params.items() if k in self.param})
        self._panel = None
        for filt in self.filters:
            filt.param.watch(self.update, 'value')
        self._cache = None
        self._updates = None
        self.update()

    @classmethod
    def _get_type(cls, view_type):
        """
        Returns the matching View type.
        """
        try:
            __import__(f'lumen.views.{view_type}')
        except Exception:
            pass
        for view in param.concrete_descendents(cls).values():
            if view.view_type == view_type:
                return view
        if view_type is not None:
            raise ValueError(f"View type '{view_type}' could not be found.")
        return View

    @classmethod
    def from_spec(cls, spec, monitor, source, filters):
        """
        Resolves a View specification given the schema of the Source
        it will be filtering on.

        Parameters
        ----------
        spec: dict
            Specification declared as a dictionary of parameter values.
        monitor: lumen.monitor.Monitor
            The Monitor object that holds this View.
        source: lumen.sources.Source
            The Source object containing the tables the View renders.
        filters: list(lumen.filters.Filter)
            A list of Filter objects which provide query values for
            the Source.

        Returns
        -------
        The resolved View object.
        """
        spec = dict(spec)
        transform_specs = spec.pop('transforms', [])
        transforms = [Transform.from_spec(tspec) for tspec in transform_specs]
        view_type = View._get_type(spec.pop('type', None))
        return view_type(
            filters=filters, monitor=monitor, source=source,
            transforms=transforms, **spec
        )

    def __bool__(self):
        return self._cache is not None and len(self._cache) > 0

    def _update_panel(self, *events):
        """
        Updates the cached Panel object and notifies the containing
        Monitor object if it is stale and has to rerender.
        """
        if self._panel is not None:
            self._updates = self._get_params()
            if self._updates is not None:
                return False
        self._panel = self.get_panel()
        return True

    def get_data(self):
        """
        Queries the Source for the specified table applying any
        filters and transformations specified on the View. Unlike
        `get_value` this should be used when multiple return values
        are expected.

        Returns
        -------
        DataFrame
            The queried table after filtering and transformations are
            applied.
        """
        if self._cache is not None:
            return self._cache
        query = {
            filt.field: filt.query for filt in self.filters
            if filt.query is not None and
            (filt.table is None or filt.table == self.table)
        }
        data = self.source.get(self.table, **query)
        for transform in self.transforms:
            data = transform.apply(data)
        if len(data):
            data = self.source._filter_dataframe(data, **query)
        self._cache = data
        return data

    def get_value(self, field=None):
        """
        Queries the Source for the data associated with a particular
        field applying any filters and transformations specified on
        the View. Unlike `get_data` this method returns a single
        scalar value associated with the field and should therefore
        only be used if only a single.

        Parameters
        ----------
        field: str (optional)
            The field from the table to return; if None uses
            field defined on the View.

        Returns
        -------
        object
            A single scalar value representing the current value of
            the queried field.
        """
        data = self.get_data()
        if not len(data) or field is not None and field not in data.columns:
            return None
        row = data.iloc[-1]
        return row[self.field if field is None else field]

    def get_panel(self):
        """
        Constructs and returns a Panel object which will represent a
        view of the queried table.

        Returns
        -------
        panel.Viewable
            A Panel Viewable object representing a current
            representation of the queried table.
        """
        return pn.panel(self.get_data())

    def update(self, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        invalidate_cache : bool
            Whether to force a rerender of the containing Monitor.

        Returns
        -------
        stale : bool
            Whether the panel on the View is stale and needs to be
            rerendered.
        """
        if invalidate_cache:
            self._cache = None
        return self._update_panel()

    def _get_params(self):
        return None

    @property
    def panel(self):
        return self._panel


class StringView(View):
    """
    The StringView renders the latest value of the field as a HTML
    string with a specified fontsize.
    """

    font_size = param.String(default='24pt', doc="""
        The font size of the rendered field value.""")

    view_type = 'string'

    def get_panel(self):
        return pn.pane.HTML(**self._get_params())

    def _get_params(self):
        value = self.get_value()
        params = dict(self.kwargs)
        if value is None:
            params['object'] = 'No info'
        else:
            params['object'] = f'<p style="font-size: {self.font_size}">{value}</p>'
        return params


class IndicatorView(View):
    """
    The IndicatorView renders the latest field value as a Panel
    Indicator.
    """

    indicator = param.String(doc="The name of the panel Indicator type")

    label = param.String(doc="""
        A custom label to use for the Indicator.""")

    view_type = 'indicator'

    def __init__(self, **params):
        super().__init__(**params)
        name = params.get('label', params.get('field', ''))
        self.kwargs['name'] = name
        self._panel.name = name

    def get_panel(self):
        indicators = param.concrete_descendents(pn.widgets.indicators.ValueIndicator)
        indicator_name = self.indicator.title()
        indicator = indicators.get(indicator_name)
        if indicator is None:
            raise ValueError(f'No such indicator as {self.indicator}, '
                             'ensure the indicator option in the spec '
                             'specifies a Panel ValueIndicator that '
                             'exists and has been imported')
        return indicator(**self._get_params())

    def _get_params(self):
        value = self.get_value()
        if (not isinstance(value, (type(None), str)) and np.isnan(value)):
            value = None
        return dict(self.kwargs, value=value)


class hvPlotView(View):
    """
    The hvPlotView renders the queried data as a bokeh plot generated
    with hvPlot. hvPlot allows for a concise declaration of a plot via
    its simple API.
    """

    kind = param.String(doc="The kind of plot, e.g. 'scatter' or 'line'.")

    x = param.String(doc="The column to render on the x-axis.")

    y = param.String(doc="The column to render on the y-axis.")

    opts = param.Dict(default={}, doc="HoloViews option to apply on the plot.")

    streaming = param.Boolean(default=False, doc="""
      Whether to stream new data to the plot or rerender the plot.""")

    view_type = 'hvplot'

    def __init__(self, **params):
        import hvplot.pandas # noqa
        self._stream = None
        super().__init__(**params)

    def get_plot(self, df):
        processed = {}
        for k, v in self.kwargs.items():
            if k.endswith('formatter') and isinstance(v, str) and '%' not in v:
                v = NumeralTickFormatter(format=v)
            processed[k] = v
        if self.streaming:
            processed['stream'] = self._stream
        plot = df.hvplot(
            kind=self.kind, x=self.x, y=self.y, **processed
        )
        return plot.opts(**self.opts) if self.opts else plot

    def get_panel(self):
        return pn.pane.HoloViews(**self._get_params())

    def _get_params(self):
        df = self.get_data()
        if self.streaming:
            from holoviews.streams import Pipe
            self._stream = Pipe(data=df)
        return dict(object=self.get_plot(df))

    def update(self, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        invalidate_cache : bool
            Whether to force a rerender of the containing Monitor.

        Returns
        -------
        stale : bool
            Whether the panel on the View is stale and needs to be
            rerendered.
        """
        if invalidate_cache:
            self._cache = None
        if not self.streaming or self._stream is None:
            return self._update_panel()
        self._stream.send(self.get_data())
        return False


class Table(View):
    """
    Renders a Source table using a Panel Table widget.
    """

    view_type = 'table'

    def get_panel(self):
        return pn.widgets.tables.Tabulator(**self._get_params())

    def _get_params(self):
        return dict(value=self.get_data(), disabled=True, **self.kwargs)


class Download(View):
    """
    The Download View allows downloading the current table as a csv or
    xlsx file.
    """

    filename = param.String(default='data', doc="""
      Filename of the downloaded file.""")

    filetype = param.Selector(default='csv', objects=['csv', 'xlsx'], doc="""
      File type of the downloaded file.""")

    save_kwargs = param.Dict(default={}, doc="""
      Options for the to_csv or to_excel methods.""")

    view_type = 'download'

    def __bool__(self):
        return True

    def _get_file_data(self):
        df = self.get_data()
        sio = StringIO()
        if self.filetype == 'csv':
            savefn = df.to_csv
        elif self.filetype == 'xlsx':
            savefn = df.to_excel
        savefn(sio, **self.save_kwargs)
        sio.seek(0)
        return sio

    def get_panel(self):
        return pn.widgets.FileDownload(**self._get_params())

    def _get_params(self):
        filename = f'{self.filename}.{self.filetype}'
        return dict(filename=filename, callback=self._get_file_data, **self.kwargs)



import datetime as dt

from functools import partial
from itertools import product

import param
import panel as pn

from .filters import FacetFilter
from .sources import Source
from .views import View


class Monitor(param.Parameterized):
    """
    A Monitor renders the results of a Source query using the defined
    set of filters and views.
    """

    application = param.Parameter(doc="The overall monitoring application.")

    filters = param.List(doc="A list of filters to be rendered.")

    layout = param.ClassSelector(default='column', class_=(str, list), doc="""
        Defines the layout of the views in the monitor target. Can be
        'column', 'row', 'grid', 'row' or a nested list of indexes
        corresponding to the views, e.g. [[0, 1], [2]] will create
        a Column of one row containing views 0 and 1 and a second Row
        containing view 2.""" )

    sort_fields = param.List(default=[], doc="List of fields to sort by.")

    sort_reverse = param.Boolean(default=False, doc="""
        Whether to reverse the sort order.""")

    schema = param.Dict()

    title = param.String(doc="A title for this monitor.")

    refresh_rate = param.Integer(default=None, doc="""
        How frequently to refresh the monitor by querying the adaptor.""")

    source = param.ClassSelector(class_=Source, doc="""
       The Source queries the data from some data source.""")

    tsformat = param.String(default="%m/%d/%Y %H:%M:%S")

    views = param.List(doc="A list of views to be displayed.")

    def __init__(self, **params):
        sort = params.pop('sort', {})
        params['sort_fields'] = sort_fields = sort.get('fields', [])
        params['sort_reverse'] = sort_reverse = sort.get('reverse', False)
        self._sort_widget = pn.widgets.MultiSelect(
            options=sort_fields,
            sizing_mode='stretch_width',
            size=len(sort_fields),
            value=sort_fields,
        )
        self._reverse_widget = pn.widgets.Checkbox(
            value=sort_reverse, name='Reverse', margin=(5, 0, 0, 10)
        )
        self._reload_button = pn.widgets.Button(
            name='↻', width=50, css_classes=['reload'], margin=0
        )
        self._reload_button.on_click(self.update)
        self._config = params.pop('config', {})
        self.kwargs = {k: v for k, v in params.items() if k not in self.param}
        super().__init__(**{k: v for k, v in params.items() if k in self.param})
        self._cards = []
        self._cache = {}
        self._cb = None
        self._stale = False
        self.timestamp = pn.pane.HTML(
            f'Last updated: {dt.datetime.now().strftime(self.tsformat)}',
            align='end', margin=10, sizing_mode='stretch_width'
        )
        self._updates = {}
        self._update_views()
        for filt in self.filters:
            if isinstance(filt, FacetFilter):
                continue
            filt.param.watch(partial(self._rerender, invalidate_cache=True), 'value')

    @pn.depends('_sort_widget.value', '_reverse_widget.value', watch=True)
    def _resort(self, *events):
        self.sort_fields = self._sort_widget.value
        self.sort_reverse = self._reverse_widget.value
        self._rerender(update_views=False)

    def _get_sort_key(self, views):
        sort_key = []
        for field in self.sort_fields:
            values = [v.get_value(field) for v in views]
            if values:
                sort_key.append(values[0])
        return tuple(sort_key)

    def _construct_card(self, title, views):
        kwargs = dict(self.kwargs)
        if isinstance(self.layout, list):
            item = pn.Column(sizing_mode='stretch_both')
            for row_spec in self.layout:
                row = pn.Row(sizing_mode='stretch_both')
                for index in row_spec:
                    row.append(views[index].panel)
                item.append(row)
        else:
            if self.layout == 'grid' and 'ncols' not in kwargs:
                kwargs['ncols'] = 2
            layout = {
                'column': pn.Column,
                'grid': pn.GridBox,
                'row': pn.Row,
                'tabs': pn.Tabs
            }[self.layout]
            item = layout(*(view.panel for view in views), **kwargs)

        params = {k: v for k, v in self.kwargs.items() if k in pn.Card.param}
        return pn.Card(
            item, title=title, name=title, **params
        )

    def _get_card(self, filters, facet_filters, invalidate_cache=True, update_views=True):
        view_filters = filters + list(facet_filters)
        key = (tuple(str(f.value) for f in facet_filters) +
               tuple(id(view) for view in self.views))

        update_card = False
        if key in self._cache:
            card, views = self._cache[key]
            if update_views:
                for view in views:
                    view_stale = view.update(invalidate_cache)
                    update_card = update_card or view_stale
        else:
            card = None
            views = [
                View.from_spec(view_spec, self, self.source, view_filters)
                for view_spec in self.views
            ]
        if not any(view for view in views):
            return None, None

        sort_key = self._get_sort_key(views)

        if facet_filters:
            title = ' '.join([f'{f.label}: {f.value}' for f in facet_filters])
        else:
            title = self.title

        if card is None:
            card = self._construct_card(title, views)
            self._cache[key] = (card, views)
        else:
            card.title = title
            if update_card:
                self._updates[card] = views
        return sort_key, card

    def _update_views(self, invalidate_cache=True, update_views=True):
        filters = [filt for filt in self.filters
                   if not isinstance(filt, FacetFilter)]
        facets = [filt.filters for filt in self.filters
                  if isinstance(filt, FacetFilter)]

        cards = []
        for facet_filters in product(*facets):
            key, card = self._get_card(
                filters, facet_filters, invalidate_cache, update_views
            )
            if card is None:
                continue
            cards.append((key, card))

        if self.sort_fields:
            cards = sorted(cards, key=lambda x: x[0])
            if self.sort_reverse:
                cards = cards[::-1]

        cards = [card for _, card in cards]
        if cards != self._cards:
            self._cards[:] = cards
            self._stale = True

    def _rerender(self, *events, invalidate_cache=False, update_views=True):
        self._update_views(invalidate_cache, update_views)
        has_updates = (
            any(view._updates for _, (_, views) in self._cache.items() for view in views) or
            bool(self._updates)
        )
        if update_views and has_updates:
            self.application._loading(self.title)
            for card, views in self._updates.items():
                card[0][:] = [view.panel for view in views]
            self._updates = {}
            for _, (_, views) in self._cache.items():
                for view in views:
                    if view._updates:
                        view._panel.param.set_param(**view._updates)
                        view._updates = None
        if self._stale or (update_views and has_updates):
            self.application._rerender()
            self._stale = False

    # Public API

    @property
    def filter_panel(self):
        views = []
        source_panel = self.source.panel
        if source_panel:
            source_header = pn.pane.Markdown('### Source', margin=(0, 5))
            views.extend([source_header, source_panel, pn.layout.Divider()])
        filters = [filt.panel for filt in self.filters if filt.panel is not None]
        if filters:
            views.append(pn.pane.Markdown('### Filters', margin=(0, 5)))
            views.extend(filters)
            views.append(pn.layout.Divider())
        if self.sort_fields:
            views.extend([
                pn.pane.Markdown('### Sort', margin=(0, 5)),
                self._sort_widget,
                self._reverse_widget
            ])
            views.append(pn.layout.Divider())
        views.append(pn.Row(self._reload_button, self.timestamp, sizing_mode='stretch_width'))
        return pn.Column(*views, name=self.title, sizing_mode='stretch_width') if views else None

    @property
    def panels(self):
        default = 'grid' if len(self._cards) > 1 else 'column'
        layout = self._config.get('layout', default)
        kwargs = dict(name=self.title, sizing_mode='stretch_width')
        if layout == 'grid':
            kwargs['ncols'] = self._config.get('ncols', 3)
            return pn.GridBox(*self._cards, **kwargs)
        elif layout == 'column':
            return pn.Column(*self._cards, **kwargs)

    @pn.depends('refresh_rate', watch=True)
    def start(self, event=None):
        refresh_rate = self.refresh_rate if event is None else event.new
        if refresh_rate is None:
            return
        if self._cb:
            self._cb.period = refresh_rate
        else:
            self._cb = pn.state.add_periodic_callback(
                self.update, refresh_rate
            )

    def update(self, *events):
        self.source.clear_cache()
        self.timestamp.object = f'Last updated: {dt.datetime.now().strftime(self.tsformat)}'
        self._rerender(invalidate_cache=True)



import os
import yaml
import importlib.util

import param
import panel as pn

from panel.template.base import BasicTemplate
from panel.template import DefaultTheme, DarkTheme

from .filters import ConstantFilter, Filter, WidgetFilter # noqa
from .monitor import Monitor # noqa
from .sources import Source, RESTSource # noqa
from .transforms import Transform # noqa
from .util import expand_spec
from .views import View # noqa

_templates = {k[:-8].lower(): v for k, v in param.concrete_descendents(BasicTemplate).items()}

_THEMES = {'default': DefaultTheme, 'dark': DarkTheme}

pn.config.raw_css.append("""
.reload .bk-btn {
  background: transparent;
  border: none;
  font-size: 18pt;
}
#header .reload .bk-btn {
  color: white;
}
.reload .bk-btn:hover {
  background: transparent;
}
""")

class Dashboard(param.Parameterized):

    specification = param.Filename()

    def __init__(self, specification, **params):
        super(Dashboard, self).__init__(
            specification=specification, **params
        )
        self._load_config(from_file=True)
        self._modules = {}
        self._load_local_modules()

        # Construct template
        tmpl = self.config.get('template', 'material')
        theme = self.config.get('theme', 'default').lower()
        kwargs = {'title': self.config.get('title', 'Lumen Dashboard'),
                  'theme': _THEMES[theme]}
        if 'logo' in self.config:
            kwargs['logo'] = self.config['logo']
        self.template = _templates[tmpl](**kwargs)

        # Add editor modal
        self._edited = False
        self._editor = pn.widgets.Ace(
            value=self._yaml, filename=self.specification,
            sizing_mode='stretch_both', min_height=600,
            theme='monokai'
        )
        self._edit_button = pn.widgets.Button(
            name='✎', width=50, css_classes=['reload'], margin=0,
            align='center'
        )
        self._editor.param.watch(self._edit, 'value')
        self._edit_button.on_click(self._open_modal)
        self._editor_layout = pn.Column(
            f'## Edit {os.path.basename(self.specification)}',
            self._editor, sizing_mode='stretch_both'
        )
        self.template.modal.append(self._editor_layout)

        # Build layouts
        self.filters = pn.Accordion(margin=0, sizing_mode='stretch_width')
        layout = self.config.get('layout', 'grid')
        if layout == 'tabs':
            self.targets = pn.Tabs(sizing_mode='stretch_width', dynamic=True)
            self.targets.param.watch(self._activate_filters, 'active')
        elif layout == 'column':
            self.targets = pn.Column(margin=10, sizing_mode='stretch_width')
        elif layout == 'grid':
            ncols = self.config.get('ncols', 3)
            self.targets = pn.GridBox(margin=10, ncols=ncols,
                                      sizing_mode='stretch_width')
        self._reload_button = pn.widgets.Button(
            name='↻', width=50, css_classes=['reload'], margin=0,
            align='center'
        )
        self._reload_button.on_click(self._reload)
        self._reload()

        # Populate template
        if len(self.filters):
            self.template.sidebar[:] = [self.filters]
        self.template.main[:] = [self.targets]
        header = pn.Row(self._reload_button, self._edit_button)
        self.template.header.append(header)
        if 'auth' in self._spec:
            header.extend([
                pn.layout.HSpacer(),
                f'<b><font size="4.5em">User: {pn.state.user}</font></b>'
            ])

    @property
    def _authorized(self):
        authorized = True
        for k, value in self._spec.get('auth', {}).items():
            if not isinstance(value, list): value = [value]
            if k in pn.state.user_info:
                user_value = pn.state.user_info[k]
                if not isinstance(user_value, list):
                    user_value = [user_value]
                authorized &= any(uv == v for v in value for uv in user_value)
        return authorized

    def _edit(self, event):
        self._yaml = event.new
        self._edited = True

    def _activate_filters(self, event):
        self.filters.active = [event.new]

    def _open_modal(self, event):
        # Workaround until Panel https://github.com/holoviz/panel/pull/1719
        # is released
        self.template.close_modal()
        self.template.open_modal()

    def _load_config(self, from_file=False):
        # Load config
        from . import config
        kwargs = {}
        if from_file or self._yaml is None:
            with open(self.specification) as f:
                self._yaml = f.read()
        elif self._edited:
            kwargs = {'getenv': False, 'getshell': False, 'getoauth': False}
        spec = expand_spec(self._yaml, config.template_vars, **kwargs)
        self._spec = yaml.load(spec, Loader=yaml.Loader)
        if not 'targets' in self._spec:
            raise ValueError('Yaml specification did not declare any targets.')
        self.config = self._spec.get('config', {})
        self._apply_defaults(self._spec.get('defaults', {}))
        self._root = os.path.abspath(os.path.dirname(self.specification))

    def _load_local_modules(self):
        for imp in ('filters', 'sources', 'transforms', 'views'):
            path = os.path.join(self._root, imp+'.py')
            if not os.path.isfile(path):
                continue
            spec = importlib.util.spec_from_file_location(f"local_lumen.{imp}", path)
            self._modules[imp] = module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

    def _apply_defaults(self, defaults):
        for key, obj in (('filters', Filter), ('sources', Source),
                         ('transforms', Transform), ('views', View)):
            for default in defaults.get(key, []):
                defaults = dict(default)
                obj_type = obj._get_type(defaults.pop('type', None))
                obj_type.param.set_param(**defaults)

    def _reload(self, *events):
        self._load_config()
        if not self._authorized:
            auth_keys = list(self._spec.get('auth'))
            if len(auth_keys) == 1:
                auth_keys = repr(auth_keys[0])
            else:
                auth_keys = [repr(k) for k in auth_keys]
            error = ('## Unauthorized User\n'
                     f'{pn.state.user} is not authorized. Ensure '
                     f'{pn.state.user} is permissioned correctly on '
                     f'the {auth_keys} field(s) in OAuth user data.')
            self.targets[:] = [pn.pane.Alert(error, alert_type='danger')]
            self._targets = []
            return
        self._sources = {}
        for name, source_spec in self._spec.get('sources', {}).items():
            self._sources[name] = Source.from_spec(source_spec, self._sources,
                                                   root=self._root)
        self._targets = self._resolve_targets(self._spec['targets'])
        self._rerender()
        filters = []
        for target in self._targets:
            panel = target.filter_panel
            if panel is not None:
                filters.append(panel)
        self.filters[:] = filters
        self.filters.active = [0]

    def _loading(self, name=''):
        loading = pn.Column(
            pn.indicators.LoadingSpinner(value=True, align='center'),
            f'**Reloading {name}...**'
        )
        if isinstance(self.targets, pn.GridBox):
            items = [pn.pane.HTML(width=self.targets[i].width)
                     for i in range(self.targets.ncols) if i < len(self.targets)]
            index = int(min(self.targets.ncols, (len(self.targets)-1)) / 2)
            if items:
                items[index] = loading
            else:
                items = [loading]
        elif isinstance(self.targets, pn.Tabs):
            items = list(zip(self.targets._names, list(self.targets.objects)))
            items[self.targets.active] = pn.Row(
                pn.layout.HSpacer(), loading, pn.layout.HSpacer()
            )
        else:
            items = [pn.Row(pn.layout.HSpacer(), loading, pn.layout.HSpacer())]
        self.targets[:] = items

    def _rerender(self):
        self.targets[:] = [target.panels for target in self._targets]

    def _resolve_targets(self, target_specs, metadata={}):
        targets = []
        for target_spec in target_specs:
            target_spec = dict(target_spec)

            # Resolve source
            source_spec = target_spec.pop('source', None)
            source = Source.from_spec(source_spec, self._sources, root=self._root)
            schema = source.get_schema()

            # Resolve filters
            filter_specs = target_spec.pop('filters', [])
            filters = [
                Filter.from_spec(filter_spec, schema)
                for filter_spec in filter_specs
            ]

            # Create target
            target = Monitor(
                application=self, filters=filters, source=source,
                schema=schema, **target_spec)
            targets.append(target)
        return targets

    def layout(self):
        """
        Returns a layout rendering the dashboard contents.
        """
        spinner = pn.indicators.LoadingSpinner(width=40, height=40)
        pn.state.sync_busy(spinner)

        title = self.config.get('title', 'Lumen Dashboard')
        title_html = '<font color="white"><h1>'+title+'</h1></font>'
        return pn.Column(
            pn.Row(
                pn.pane.HTML(title_html),
                pn.layout.HSpacer(),
                spinner,
                background='#00aa41'
            ),
            pn.Row(
                pn.Column(self.filters, width=300),
                self.targets
            )
        )

    def show(self):
        """
        Opens the dashboard in a new window.
        """
        title = self.config.get('title', 'Lumen Dashboard')
        self.template.show(title=title)
        for target in self._targets:
            target.start()

    def servable(self):
        """
        Marks the dashboard for serving with `panel serve`.
        """
        title = self.config.get('title', 'Lumen Dashboard')
        self.template.servable(title=title)
        for target in self._targets:
            target.start()



import param as _param

from .dashboard import Dashboard # noqa


class _config(_param.Parameterized):
    """
    Stores shared configuration for the entire Lumen application.
    """

    sources = _param.Dict(default={}, doc="""
      A global dictionary of shared Source objects.""")

    template_vars = _param.Dict(default={}, doc="""
      Template variables which may be referenced in a dashboard yaml
      specification.""")



config = _config()


__version__ = str(_param.version.Version(
    fpath=__file__, archive_commit="$Format:%h$", reponame="lumen"))
